# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import typing
from typing import TextIO, Union, List, Tuple

import numpy as np

def read_Ts(info: dict, numT: int = None) -> Tuple[bool, np.ndarray]:
    """
    Read temperature or inverse-temperature values from the provided info dictionary.

    Parameters
    ----------
    info : dict
        Dictionary containing temperature or inverse-temperature parameters.
    numT : int, optional
        Number of temperature or inverse-temperature values to generate (default is None).

    Returns
    -------
    as_beta : bool
        True if using inverse-temperature, False if using temperature.
    betas : np.ndarray
        Sequence of inverse-temperature values.

    Raises
    ------
    ValueError
        If numT is not specified, or if both Tmin/Tmax and bmin/bmax are defined, or if neither are defined,
        or if bmin/bmax or Tmin/Tmax values are invalid.
    RuntimeError
        If the mode is unknown (neither set_T nor set_b).
    """
    if numT is None:
        raise ValueError("read_Ts: numT is not specified")

    Tmin = info.get("Tmin", None)
    Tmax = info.get("Tmax", None)
    bmin = info.get("bmin", None)
    bmax = info.get("bmax", None)
    logscale = info.get("Tlogspace", True)

    if "Tinvspace" in info:
        raise ValueError("Tinvspace is deprecated. Use bmax/bmin instead.")

    set_b = (bmin is not None or bmax is not None)
    set_T = (Tmin is not None or Tmax is not None)

    if set_b and set_T:
        raise ValueError("both Tmin/Tmax and bmin/bmax are defined")
    if (not set_b) and (not set_T):
        raise ValueError("neither Tmin/Tmax nor bmin/bmax are defined")

    if set_b:
        if bmin is None or bmax is None:
            raise ValueError("bmin and bmax must be set")

        input_as_beta = True
        if not np.isreal(bmin) or bmin < 0.0:
            raise ValueError("bmin must be zero or a positive real number")
        if not np.isreal(bmax) or bmax < 0.0:
            raise ValueError("bmin must be zero or a positive real number")
        if bmin > bmax:
            raise ValueError("bmin must be smaller than or equal to bmax")

        if logscale:
            if bmin == 0.0:
                raise ValueError("bmin must be greater than 0.0 when Tlogspace is True")
            betas = np.logspace(start=np.log10(bmin), stop=np.log10(bmax), num=numT)
        else:
            betas = np.linspace(start=bmin, stop=bmax, num=numT)

    elif set_T:
        if Tmin is None or Tmax is None:
            raise ValueError("Tmin and Tmax must be set")

        input_as_beta = False
        if not np.isreal(Tmin) or Tmin <= 0.0:
            raise ValueError("Tmin must be a positive real number")
        if not np.isreal(Tmax) or Tmax <= 0.0:
            raise ValueError("Tmax must be a positive real number")
        if Tmin > Tmax:
            raise ValueError("Tmin must be smaller than or equal to Tmax")

        if logscale:
            Ts = np.logspace(start=np.log10(Tmin), stop=np.log10(Tmax), num=numT)
        else:
            Ts = np.linspace(start=Tmin, stop=Tmax, num=numT)

        betas = 1.0 / Ts
    else:
        raise RuntimeError("read_Ts: unknown mode: not set_T nor set_b")

    return input_as_beta, betas



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Dict, Union

import os
import numpy as np
#-- delay import
# from mpi4py.util.dtlib import from_numpy_dtype

import odatse.mpi

_use_buffer = int(os.environ.get("ODATSE_USE_MPI_BUFFERED", 0)) == 1

# def set_config(*, use_buffer):
#     global _use_buffer
#     _use_buffer = use_buffer

def gather_replica(data, axis=0):
    mpisize = odatse.mpi.size()
    if mpisize == 1:
        return data
    else:
        if _use_buffer:
            return _do_gather_replica_buffer(data, axis)
        else:
            return _do_gather_object(data, axis)

def gather_data(data, axis=0):
    mpisize = odatse.mpi.size()
    if mpisize == 1:
        return data
    else:
        if _use_buffer:
            return _do_gather_data_buffer(data, axis)
        else:
            return _do_gather_object(data, axis)

def _do_gather_object(data, axis):
    mpicomm = odatse.mpi.comm()
    return np.concatenate(mpicomm.allgather(data), axis=axis)

def _do_gather_replica_buffer(data, axis):
    if axis == 0:
        return _do_gather_variable_buffer(data)
    else:
        return _do_gather_variable_buffer_transpose(data, axis)

def _do_gather_data_buffer(data, axis):
    if axis == 0:
        return _do_gather_fixed_buffer(data)
    else:
        return _do_gather_fixed_buffer_transpose(data, axis)

def _do_gather_variable_buffer(data):
    from mpi4py.util.dtlib import from_numpy_dtype
    mpicomm = odatse.mpi.comm()
    mpisize = odatse.mpi.size()

    sh = data.shape
    nrep = np.array([sh[0]], dtype=np.int64)
    nreps = np.zeros(mpisize, dtype=np.int64)
    mpicomm.Allgather(nrep, nreps)

    displ = np.cumsum(nreps) - nreps
    nrep_total = np.sum(nreps)
    ndim = np.prod(sh[1:], dtype=int)

    buf = np.zeros((nrep_total, *sh[1:]), dtype=data.dtype)
    dtype = from_numpy_dtype(data.dtype)
    mpicomm.Allgatherv([data, dtype], [buf, nreps*ndim, displ*ndim, dtype])
    return buf

def _do_gather_variable_buffer_transpose(data, axis):
    axes = [i for i in range(data.ndim)]
    axes[axis], axes[0] = axes[0], axes[axis]
    axes = tuple(axes)
    data_t = np.transpose(data, axes).copy()
    data_gather = _do_gather_variable_buffer(data_t)
    return np.transpose(data_gather, axes)
        
def _do_gather_fixed_buffer(data):
    from mpi4py.util.dtlib import from_numpy_dtype
    mpicomm = odatse.mpi.comm()
    mpisize = odatse.mpi.size()

    sh = data.shape
    buf = np.zeros((mpisize*sh[0], *sh[1:]), dtype=data.dtype)
    mpicomm.Allgather(data, buf)
    return buf
    
def _do_gather_fixed_buffer_transpose(data, axis):    
    axes = [i for i in range(data.ndim)]
    axes[axis], axes[0] = axes[0], axes[axis]
    axes = tuple(axes)
    data_t = np.transpose(data, axes).copy()
    data_gather = _do_gather_fixed_buffer(data_t)
    return np.transpose(data_gather, axes)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Union, Dict, Optional

from pathlib import Path
from io import open
import numpy as np
import os
import time

import odatse
import odatse.domain

from .mapper_mpi_base import Algorithm as MapperMPIAlgorithm

class Algorithm(MapperMPIAlgorithm):
    """
    Algorithm class for data analysis of quantum beam diffraction experiments.
    Inherits from odatse.algorithm.mapper_mpi.Algorithm.
    """
    mesh_list: List[Union[int, float]]

    def __init__(self, info: odatse.Info,
                 runner: Optional[odatse.Runner] = None,
                 domain = None,
                 run_mode: str = "initial"
    ) -> None:
        """
        Initialize the Algorithm instance.

        Parameters
        ----------
        info : Info
            Information object containing algorithm parameters.
        runner : Runner
            Optional runner object for submitting tasks.
        domain :
            Optional domain object, defaults to MeshGrid.
        run_mode : str
            Mode to run the algorithm, defaults to "initial".
        """

        super().__init__(info=info, runner=runner, domain=domain, run_mode=run_mode, meshgrid=False)



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import List, Union, Dict, Optional

from pathlib import Path
from io import open
import numpy as np
import os
import time

import odatse
import odatse.domain
from ._algorithm import AlgorithmBase

class Algorithm(AlgorithmBase):
    """
    Algorithm class for data analysis of quantum beam diffraction experiments.
    Inherits from odatse.algorithm.AlgorithmBase.
    """
    mesh_list: List[Union[int, float]]

    def __init__(self, info: odatse.Info,
                 runner: Optional[odatse.Runner] = None,
                 domain = None,
                 run_mode: str = "initial",
                 meshgrid: bool = True,
    ) -> None:
        """
        Initialize the Algorithm instance.

        Parameters
        ----------
        info : Info
            Information object containing algorithm parameters.
        runner : Runner
            Optional runner object for submitting tasks.
        domain :
            Optional domain object, defaults to MeshGrid.
        run_mode : str
            Mode to run the algorithm, defaults to "initial".
        meshgrid : bool
            Whether to use mesh grid or points.
        """
        super().__init__(info=info, runner=runner, run_mode=run_mode)

        if domain and isinstance(domain, odatse.domain.MeshGrid):
            self.domain = domain
        else:
            self.domain = odatse.domain.MeshGrid(info, rng=self.rng, mesh=meshgrid)

        self.domain.do_split()
        self.mesh_list = self.domain.grid_local

        self.colormap_file = info.algorithm.get("colormap", "ColorMap.txt")
        self.local_colormap_file = Path(self.colormap_file).name + ".tmp"

    def _initialize(self) -> None:
        """
        Initialize the algorithm parameters and timer.
        """
        self.fx_list = []
        self.timer["run"]["submit"] = 0.0
        self._show_parameters()

    def _run(self) -> None:
        """
        Execute the main algorithm process.
        """
        # Make ColorMap

        if self.mode is None:
            raise RuntimeError("mode unset")

        if self.mode.startswith("init"):
            self._initialize()
        elif self.mode.startswith("resume"):
            self._load_state(self.checkpoint_file)
        else:
            raise RuntimeError("unknown mode {}".format(self.mode))

        # local colormap file
        fp = open(self.local_colormap_file, "a")
        if self.mode.startswith("init"):
            fp.write("#" + " ".join(self.label_list) + " fval\n")

        iterations = len(self.mesh_list)
        istart = len(self.fx_list)

        next_checkpoint_step = istart + self.checkpoint_steps
        next_checkpoint_time = time.time() + self.checkpoint_interval

        for icount in range(istart, iterations):
            print("Iteration : {}/{}".format(icount+1, iterations))
            mesh = self.mesh_list[icount]

            # update information
            args = (int(mesh[0]), 0)
            x = np.array(mesh[1:])

            time_sta = time.perf_counter()
            fx = self.runner.submit(x, args)
            time_end = time.perf_counter()
            self.timer["run"]["submit"] += time_end - time_sta

            self.fx_list.append([mesh[0], fx])

            # write to local colormap file
            fp.write(" ".join(
                map(lambda v: "{:8f}".format(v), (*x, fx))
            ) + "\n")

            if self.checkpoint:
                time_now = time.time()
                if icount+1 >= next_checkpoint_step or time_now >= next_checkpoint_time:
                    self._save_state(self.checkpoint_file)
                    next_checkpoint_step = icount + 1 + self.checkpoint_steps
                    next_checkpoint_time = time_now + self.checkpoint_interval

        if iterations > 0:
            opt_index = np.argsort(self.fx_list, axis=0)[0][1]
            opt_id, opt_fx = self.fx_list[opt_index]
            opt_mesh = self.mesh_list[opt_index]

            self.opt_fx = opt_fx
            self.opt_mesh = opt_mesh

            print(f"[{self.mpirank}] minimum_value: {opt_fx:12.8e} at {opt_mesh[1:]} (mesh {opt_mesh[0]})")

        self._output_results()

        if Path(self.local_colormap_file).exists():
            os.remove(Path(self.local_colormap_file))

        print("complete main process : rank {:08d}/{:08d}".format(self.mpirank, self.mpisize))

    def _output_results(self):
        """
        Output the results to the colormap file.
        """
        print("Make ColorMap")
        time_sta = time.perf_counter()

        with open(self.colormap_file, "w") as fp:
            fp.write("#" + " ".join(self.label_list) + " fval\n")

            for x, (idx, fx) in zip(self.mesh_list, self.fx_list):
                fp.write(" ".join(
                    map(lambda v: "{:8f}".format(v), (*x[1:], fx))
                    ) + "\n")

            if len(self.mesh_list) > 0:
                fp.write("#Minimum point : " + " ".join(
                    map(lambda v: "{:8f}".format(v), self.opt_mesh[1:])
                ) + "\n")
                fp.write("#R-factor : {:8f}\n".format(self.opt_fx))
                fp.write("#see Log{:d}\n".format(round(self.opt_mesh[0])))
            else:
                fp.write("# No mesh point\n")

        time_end = time.perf_counter()
        self.timer["run"]["file_CM"] = time_end - time_sta

    def _prepare(self) -> None:
        """
        Prepare the algorithm (no operation).
        """
        pass

    def _post(self) -> Dict:
        """
        Post-process the results and gather data from all MPI ranks.

        Returns
        -------
        Dict
            Dictionary of results.
        """
        if self.mpisize > 1:
            fx_lists = self.mpicomm.allgather(self.fx_list)
            results = [v for vs in fx_lists for v in vs]
        else:
            results = self.fx_list

        if self.mpirank == 0:
            with open(self.colormap_file, "w") as fp:
                for x, (idx, fx) in zip(self.domain.grid, results):
                    assert x[0] == idx
                    fp.write(" ".join(
                        map(lambda v: "{:8f}".format(v), (*x[1:], fx))
                    ) + "\n")

        return {}

    def _save_state(self, filename) -> None:
        """
        Save the current state of the algorithm to a file.

        Parameters
        ----------
        filename
            The name of the file to save the state to.
        """
        data = {
            "mpisize": self.mpisize,
            "mpirank": self.mpirank,
            "timer": self.timer,
            "info": self.info,
            "fx_list": self.fx_list,
            "mesh_size": len(self.mesh_list),
        }
        self._save_data(data, filename)

    def _load_state(self, filename, restore_rng=True):
        """
        Load the state of the algorithm from a file.

        Parameters
        ----------
        filename
            The name of the file to load the state from.
        restore_rng : bool
            Whether to restore the random number generator state.
        """
        data = self._load_data(filename)
        if not data:
            print("ERROR: Load status file failed")
            sys.exit(1)

        assert self.mpisize == data["mpisize"]
        assert self.mpirank == data["mpirank"]

        self.timer = data["timer"]

        info = data["info"]
        self._check_parameters(info)

        self.fx_list = data["fx_list"]

        assert len(self.mesh_list) == data["mesh_size"]



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import numpy as np
import odatse
import time

# type hints
from pathlib import Path
from typing import Callable, Optional, Dict, Tuple


class Solver(odatse.solver.SolverBase):
    """
    Solver class for evaluating functions with given parameters.
    """
    x: np.ndarray
    fx: float
    _func: Optional[Callable[[np.ndarray], float]]

    def __init__(self, info: odatse.Info) -> None:
        """
        Initialize the solver.

        Parameters
        ----------
        info: Info
            Information object containing solver configuration.
        """
        super().__init__(info)
        self._name = "function"
        self._func = None

        # for debug purpose
        self.delay = info.solver.get("delay", 0.0)

    def evaluate(self, x: np.ndarray, args: Tuple = (), nprocs: int = 1, nthreads: int = 1) -> float:
        """
        Evaluate the function with given parameters.

        Parameters
        ----------
        x : np.ndarray
            Input array for the function.
        args : Tuple, optional
            Additional arguments for the function.
        nprocs : int, optional
            Number of processes to use.
        nthreads : int, optional
            Number of threads to use.

        Returns
        -------
        float
            Result of the function evaluation.
        """
        self.prepare(x, args)
        cwd = os.getcwd()
        os.chdir(self.work_dir)
        self.run(nprocs, nthreads)
        os.chdir(cwd)
        result = self.get_results()
        return result

    def prepare(self, x: np.ndarray, args = ()) -> None:
        """
        Prepare the solver with the given parameters.

        Parameters
        ----------
        x : np.ndarray
            Input array for the function.
        args : tuple, optional
            Additional arguments for the function.
        """
        self.x = x

    def run(self, nprocs: int = 1, nthreads: int = 1) -> None:
        """
        Run the function evaluation.

        Parameters
        ----------
        nprocs : int, optional
            Number of processes to use.
        nthreads : int, optional
            Number of threads to use.

        Raises
        ------
        RuntimeError
            If the function is not set.
        """
        if self._func is None:
            raise RuntimeError(
                "ERROR: function is not set. Make sure that `set_function` is called."
            )
        self.fx = self._func(self.x)
        # for debug purpose
        if self.delay > 0.0:
            time.sleep(self.delay)

    def get_results(self) -> float:
        """
        Get the results of the function evaluation.

        Returns
        -------
        float
            Result of the function evaluation.
        """
        return self.fx

    def set_function(self, f: Callable[[np.ndarray], float]) -> None:
        """
        Set the function to be evaluated.

        Parameters
        ----------
        f : Callable[[np.ndarray], float]
            Function to be evaluated.
        """
        self._func = f



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Dict, List, Optional
import pathlib
from os import PathLike
from collections import namedtuple

import numpy as np

from odatse import mpi

Entry = namedtuple("Entry", ["step", "walker", "fx", "xs"])


def separateT(
    Ts: np.ndarray,
    nwalkers: int,
    output_dir: PathLike,
    comm: Optional[mpi.Comm],
    use_beta: bool,
    buffer_size: int = 10000,
) -> None:
    """
    Separates and processes temperature data for quantum beam diffraction experiments.

    Parameters
    ----------
    Ts : np.ndarray
        Array of temperature values.
    nwalkers : int
        Number of walkers.
    output_dir : PathLike
        Directory to store the output files.
    comm : mpi.Comm, optional
        MPI communicator for parallel processing.
    use_beta : bool
        Flag to determine if beta values are used instead of temperature.
    buffer_size : int, optional
        Size of the buffer for reading input data. Default is 10000.
    """
    if comm is None:
        mpisize = 1
        mpirank = 0
    else:
        mpisize = comm.size
        mpirank = comm.rank
    buffer_size = int(np.ceil(buffer_size / nwalkers)) * nwalkers
    output_dir = pathlib.Path(output_dir)
    proc_dir = output_dir / str(mpirank)

    T2idx = {T: i for i, T in enumerate(Ts)}
    T2rank = {}
    results = []
    for rank, Ts_local in enumerate(np.array_split(Ts, mpisize)):
        d: Dict[str, List[Entry]] = {}
        for T in Ts_local:
            T2rank[str(T)] = rank
            d[str(T)] = []
        results.append(d)

    # write file header
    for T in Ts[mpirank * nwalkers : (mpirank + 1) * nwalkers]:
        idx = T2idx[T]
        with open(output_dir / f"result_T{idx}.txt", "w") as f_out:
            if use_beta:
                f_out.write(f"# beta = {T}\n")
            else:
                f_out.write(f"# T = {T}\n")

    f_in = open(proc_dir / "result.txt")
    EOF = False
    while not EOF:
        for i in range(len(results)):
            for key in results[i].keys():
                results[i][key] = []
        for _ in range(buffer_size):
            line = f_in.readline()
            if line == "":
                EOF = True
                break
            line = line.split("#")[0].strip()
            if len(line) == 0:
                continue
            words = line.split()
            step = int(words[0])
            walker = mpirank * nwalkers + int(words[1])
            Tstr = words[2]
            fx = words[3]
            xs = words[4:]
            entry = Entry(step=step, walker=walker, fx=fx, xs=xs)
            rank = T2rank[Tstr]
            results[rank][Tstr].append(entry)
        if mpisize > 1:
            results2 = comm.alltoall(results)
        else:
            results2 = results
        d = results2[0]
        for i in range(1, len(results2)):
            for key in d.keys():
                d[key].extend(results2[i][key])
        for T in Ts[mpirank * nwalkers : (mpirank + 1) * nwalkers]:
            idx = T2idx[T]
            d[str(T)].sort(key=lambda e: e.step)
            with open(output_dir / f"result_T{idx}.txt", "a") as f_out:
                for e in d[str(T)]:
                    f_out.write(f"{e.step} ")
                    f_out.write(f"{e.walker} ")
                    f_out.write(f"{e.fx} ")
                    for x in e.xs:
                        f_out.write(f"{x} ")
                    f_out.write("\n")


def calculate_statistics_from_separated_files(
    Ts: np.ndarray,
    output_dir: PathLike,
    thermalization_steps: int,
    comm: Optional[mpi.Comm],
) -> None:
    """
    Calculate and save statistical quantities (means and errors of f(x) and partition function) from separated files generated by separateT.

    This function reads the separated files, ``result_T<Tindex>.txt`` in ``output_dir``, generated by separateT.
    The output file is ``fx.txt`` in ``output_dir``. The format is described as a header as follows:

    .. code-block::

        # $1: 1/T
        # $2: mean of f(x)
        # $3: standard error of f(x)
        # $4: number of replicas [Not used for exchange MC]
        # $5: log(Z/Z0)
        # $6: acceptance ratio

    Parameters
    ----------
    Ts : np.ndarray
        Array of temperature values.
    output_dir : PathLike
        Directory to store the output files.
    thermalization_steps : int
        Number of steps to discard for thermalization.
    comm : mpi.Comm, optional
        MPI communicator for parallel processing.
    """

    if comm is None:
        mpisize = 1
        mpirank = 0
    else:
        mpisize = comm.size
        mpirank = comm.rank
    output_dir = pathlib.Path(output_dir)

    numT = len(Ts)
    T_is_ascending = Ts[0] < Ts[1] if numT > 1 else True
    local_Tindices = np.array_split(np.arange(numT), mpisize)[mpirank]

    fx_means = np.zeros(numT)
    fx_errors = np.zeros(numT)
    dlogZs = np.zeros(numT)
    acceptances = np.zeros(numT)

    use_log_sum = True

    for Tindex in local_Tindices:
        T = Ts[Tindex]

        if T_is_ascending:
            lowerTindex = Tindex - 1
        else:
            lowerTindex = Tindex + 1
        if 0 <= lowerTindex < numT:
            lowerT = Ts[lowerTindex]
            dbeta = 1.0 / lowerT - 1.0 / T
        else:
            lowerTindex = -1
            dbeta = 0.0
        dlogZ = 0.0
        f_base = None
        fx_sum = 0.0
        fx_sum2 = 0.0
        accepted = 0
        old_x = None
        N = 0
        with open(output_dir / f"result_T{Tindex}.txt", "r") as f_in:
            for line in f_in:
                line = line.split("#")[0].strip()
                if len(line) == 0:
                    continue
                words = line.split()
                step = int(words[0])
                fx = float(words[2])
                x = np.array([float(x) for x in words[3:]])
                if step < thermalization_steps:
                    old_x = x
                    continue

                N += 1
                if not np.allclose(x, old_x):
                    accepted += 1
                old_x = x

                fx_sum += fx
                fx_sum2 += fx*fx

                if lowerTindex < 0:
                    continue

                if f_base is None:
                    f_base = fx
                    continue

                if use_log_sum:
                    ## for A > B > 0,
                    ## log(A+B) = log(A) + log(1+B/A) = log(A) + log(1+exp(log(B)-log(A)))
                    logA = dlogZ
                    logB = -dbeta * (fx - f_base)
                    if logB > logA:
                        logA, logB = logB, logA
                    dlogZ = logA + np.log1p(np.exp(logB - logA))
                else:
                    dlogZ += np.exp(-dbeta * (fx - f_base))
                
        mean = fx_sum / N
        fx_means[Tindex] = mean
        err = np.sqrt((fx_sum2/N - mean**2)/(N-1))
        fx_errors[Tindex] = err
        acceptances[Tindex] = accepted / N

        if lowerTindex < 0:
            continue
        if use_log_sum:
            dlogZ -= np.log(N)
        else:
            dlogZ = np.log(dlogZ/N)
        dlogZ += -dbeta * f_base
        dlogZs[lowerTindex] = dlogZ

    if comm is not None and mpisize > 1:
        buffer = np.zeros(numT)
        comm.Allreduce(dlogZs, buffer)
        dlogZs[:] = buffer

        comm.Allreduce(fx_means, buffer)
        fx_means[:] = buffer

        comm.Allreduce(fx_errors, buffer)
        fx_errors[:] = buffer

        comm.Allreduce(acceptances, buffer)
        acceptances[:] = buffer

    if mpirank == 0:
        dlogZ = 0.0
        with open(output_dir / "fx.txt", "w") as f_out:
            f_out.write("# $1: 1/T\n")
            f_out.write("# $2: mean of f(x)\n")
            f_out.write("# $3: standard error of f(x)\n")
            f_out.write("# $4: number of replicas [Not used for exchange MC]\n")
            f_out.write("# $5: log(Z/Z0)\n")
            f_out.write("# $6: acceptance ratio\n")
            Tindices = np.arange(numT)
            if T_is_ascending:
                Tindices = Tindices[::-1]
            for Tindex in Tindices:
                dlogZ += dlogZs[Tindex]
                T = Ts[Tindex]
                f_out.write(f"{1/T}")
                f_out.write(f" {fx_means[Tindex]}")
                f_out.write(f" {fx_errors[Tindex]}")
                f_out.write(f" {numT}")
                f_out.write(f" {dlogZ}")
                f_out.write(f" {acceptances[Tindex]}")
                f_out.write(f"\n")



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import typing
from typing import TextIO, Union, List, Tuple
import copy
import time
from pathlib import Path

import numpy as np

import odatse
from odatse.util.neighborlist import load_neighbor_list, make_neighbor_list
import odatse.util.graph
import odatse.domain
from odatse import mpi

import abc

use_buffered = False


class ContinuousState:
    def __init__(self, x):
        self.x = copy.deepcopy(x)

class DiscreteState:
    def __init__(self, inode, x):
        self.inode = copy.deepcopy(inode)
        self.x = copy.deepcopy(x)


class StateSpace(abc.ABC):
    @abc.abstractmethod
    def propose(self, state):
        ...

    @abc.abstractmethod
    def choose(self, accept, new_state, old_state):
        ...

    @abc.abstractmethod
    def gather(self, state):
        ...

    @abc.abstractmethod
    def pick(self, state, index):
        ...

    def _gather_data(self, data):
        if use_buffered:
            return self._gather_data_buffer(data)
        else:
            return self._gather_data_object(data)

    def _gather_data_object(self, data):
        mpicomm = mpi.comm()
        return np.concatenate(mpicomm.allgather(data), axis=0)

    def _gather_data_buffer(self, data):
        from mpi4py.util.dtlib import from_numpy_dtype

        mpisize = mpi.size()
        mpirank = mpi.rank()
        mpicomm = mpi.comm()

        sh = data.shape
        nrep = np.array([sh[0]], dtype=np.int64)
        nreps = np.zeros(mpisize, dtype=np.int64)
        mpicomm.Allgather(nrep, nreps)

        displ = np.cumsum(nreps) - nreps
        nrep_total = np.sum(nreps)
        ndim = np.prod(sh[1:], dtype=int)

        buf = np.zeros((nrep_total, *sh[1:]), dtype=data.dtype)
        dtype = from_numpy_dtype(data.dtype)
        mpicomm.Allgatherv([data, dtype], [buf, nreps*ndim, displ*ndim, dtype])

        return buf


class ContinuousStateSpace(StateSpace):
    def __init__(self, domain, info_param, rng, limitation):
        self.domain = domain
        self.rng = rng
        self.limitation = limitation

        self.xmin = self.domain.min_list
        self.xmax = self.domain.max_list

        if "step_list" in info_param:
            self.xstep = info_param.get("step_list")
        elif "unit_list" in info_param:
            print("WARNING: unit_list is obsolete. use step_list instead")
            self.xstep = info_param.get("unit_list")
        else:
            raise ValueError("ERROR: algorighm.param.step_list not specified")

    def initialize(self, nwalkers):
        self.domain.initialize(rng=self.rng, limitation=self.limitation, num_walkers=nwalkers)
        return ContinuousState(self.domain.initial_list)

    def propose(self, state):
        nwalkers = state.x.shape[0]
        dx = self.rng.normal(size=state.x.shape) * self.xstep
        new_state = ContinuousState(state.x + dx)
        return new_state, self._check_in_range(new_state.x), None

    def _check_in_range(self, x):
        nwalkers = x.shape[0]
        in_range = ((x >= self.xmin) & (x <= self.xmax)).all(axis=1)
        in_limit = [self.limitation.judge(x[idx,:]) for idx in range(nwalkers)]
        return in_range & in_limit

    def choose(self, accept, new_state, old_state):
        _acc = np.broadcast_to(accept.reshape(-1,1), old_state.x.shape)
        x_new = np.where(_acc, new_state.x, old_state.x)
        return ContinuousState(x_new)

    def gather(self, state):
        mpisize = mpi.size()
        if mpisize > 1:
            buf = self._gather_data(state.x)
            return ContinuousState(buf)
        else:
            return ContinuousState(state.x)

    def pick(self, state, index):
        return ContinuousState(state.x[index])

class DiscreteStateSpace(StateSpace):
    def __init__(self, domain, info_param, rng):
        self.domain = domain
        self.rng = rng

        self.node_coordinates = np.array(self.domain.grid)[:, 1:]
        self.nnodes = self.node_coordinates.shape[0]
        self._setup_neighbour(info_param)

    def initialize(self, nwalkers):
        inode = self.rng.randint(self.nnodes, size=nwalkers)
        x = self.node_coordinates[inode, :]
        return DiscreteState(inode, x)

    def propose(self, state):
        nwalkers = state.inode.shape[0]
        proposed_list = [self.rng.choice(self.neighbor_list[i]) for i in state.inode]
        proposed = np.array(proposed_list, dtype=np.int64)

        x = self.node_coordinates[proposed, :]
        new_state = DiscreteState(proposed, x)

        weight = [self.ncandidates[inode_old] / self.ncandidates[inode_new] for inode_old, inode_new in zip(state.inode, proposed_list)]

        return new_state, np.full(nwalkers, True), np.array(weight)

    def choose(self, accept, new_state, old_state):
        inode_new = np.where(accept, new_state.inode, old_state.inode)
        x_new = self.node_coordinates[inode_new, :]
        return DiscreteState(inode_new, x_new)

    def _setup_neighbour(self, info_param):
        """
        Set up the neighbor list for the discrete problem.

        Parameters
        ----------
        info_param : dict
            Dictionary containing algorithm parameters, including the path to the neighbor list file.

        Raises
        ------
        ValueError
            If the neighbor list path is not specified in the parameters.
        RuntimeError
            If the transition graph made from the neighbor list is not connected or not bidirectional.
        """
        mpirank = mpi.rank()
        mpicomm = mpi.comm()

        if "mesh_path" in info_param and "neighborlist_path" in info_param:
            nn_path = Path(info_param["neighborlist_path"]).expanduser()
            if mpirank == 0:
                nnlist = load_neighbor_list(nn_path, nnodes=self.nnodes)
            else:
                nnlist = None
            self.neighbor_list = mpicomm.bcast(nnlist, root=0)
        else:
            if "radius" not in info_param:
                raise KeyError("parameter \"algorithm.param.radius\" not specified")
            radius = info_param["radius"]
            print(f"DEBUG: create neighbor list, radius={radius}")
            self.neighbor_list = make_neighbor_list(self.node_coordinates, radius=radius, comm=mpicomm)

        # checks
        if not odatse.util.graph.is_connected(self.neighbor_list):
            raise RuntimeError(
                "ERROR: The transition graph made from neighbor list is not connected."
                "\nHINT: Increase neighborhood radius."
            )
        if not odatse.util.graph.is_bidirectional(self.neighbor_list):
            raise RuntimeError(
                "ERROR: The transition graph made from neighbor list is not bidirectional."
            )

        self.ncandidates = np.array([len(ns) - 1 for ns in self.neighbor_list], dtype=np.int64)

    def gather(self, state):
        mpisize = mpi.size()
        if mpisize > 1:
            inodes = self._gather_data(state.inode)
            return DiscreteState(inodes, self.node_coordinates[inodes, :])
        else:
            return DiscreteState(state.inode, state.x)

    def pick(self, state, index):
        return DiscreteState(state.inode[index], state.x[index])




# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from abc import ABCMeta, abstractmethod

import numpy as np

import odatse
import odatse.util.read_matrix
import odatse.util.mapping
import odatse.util.limitation
from odatse.util.logger import Logger
from odatse.exception import InputError

# type hints
from pathlib import Path
from typing import List, Optional
from . import mpi


class Run(metaclass=ABCMeta):
    def __init__(self, nprocs=None, nthreads=None, comm=None):
        """
        Initialize the Run class.

        Parameters
        ----------
        nprocs : int
            Number of processes which one solver uses.
        nthreads : int
            Number of threads which one solver process uses.
        comm : MPI.Comm
            MPI Communicator.
        """
        self.nprocs = nprocs
        self.nthreads = nthreads
        self.comm = comm

    @abstractmethod
    def submit(self, solver):
        """
        Abstract method to submit a solver.

        Parameters
        ----------
        solver : object
            Solver object to be submitted.
        """
        pass


class Runner(object):
    #solver: "odatse.solver.SolverBase"
    logger: Logger

    def __init__(self,
                 solver,
                 info: Optional[odatse.Info] = None,
                 mapping = None,
                 limitation = None) -> None:
        """
        Initialize the Runner class.

        Parameters
        ----------
        solver : odatse.solver.SolverBase
            Solver object.
        info : Optional[odatse.Info]
            Information object.
        mapping : object, optional
            Mapping object.
        limitation : object, optional
            Limitation object.
        """
        self.solver = solver
        self.solver_name = solver.name
        self.logger = Logger(info)
        self.ignore_error = info.runner.get("ignore_error", False)

        if mapping is not None:
            self.mapping = mapping
        elif "mapping" in info.runner:
            info_mapping = info.runner["mapping"]
            # N.B.: only Affine mapping is supported at present
            self.mapping = odatse.util.mapping.Affine.from_dict(info_mapping)
        else:
            # trivial mapping
            self.mapping = odatse.util.mapping.TrivialMapping()

        if limitation is not None:
            self.limitation = limitation
        elif "limitation" in info.runner:
            info_limitation = info.runner["limitation"]
            self.limitation = odatse.util.limitation.Inequality.from_dict(info_limitation)
        else:
            self.limitation = odatse.util.limitation.Unlimited()

    def prepare(self, proc_dir: Path):
        """
        Prepare the logger with the given process directory.

        Parameters
        ----------
        proc_dir : Path
            Path to the process directory.
        """
        self.logger.prepare(proc_dir)

    def submit(
            self, x: np.ndarray, args = (), nprocs: int = 1, nthreads: int = 1
    ) -> float:
        """
        Submit the solver with the given parameters.

        Parameters
        ----------
        x : np.ndarray
            Input array.
        args : tuple, optional
            Additional arguments.
        nprocs : int, optional
            Number of processes.
        nthreads : int, optional
            Number of threads.

        Returns
        -------
        float
            Result of the solver evaluation.
        """
        if self.limitation.judge(x):
            xp = self.mapping(x)
            try:
                result = self.solver.evaluate(xp, args)
            except RuntimeError as err:
                if self.ignore_error:
                    result = np.nan
                else:
                    raise
        else:
            result = np.inf
        self.logger.count(x, args, result)
        return result

    def post(self) -> None:
        """
        Write the logger data.
        """
        self.logger.write()



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Dict, Optional, TYPE_CHECKING

from io import open
import copy
import time
import itertools
import sys

import numpy as np

import odatse
import odatse.algorithm.montecarlo
from odatse.util.read_ts import read_Ts
from odatse.util.separateT import separateT, calculate_statistics_from_separated_files
from odatse.util.data_writer import DataWriter


if TYPE_CHECKING:
    from mpi4py import MPI

class Algorithm(odatse.algorithm.montecarlo.AlgorithmBase):
    """
    Replica Exchange Monte Carlo (REMC) Algorithm Implementation.
    
    This class implements the Replica Exchange Monte Carlo algorithm, also known as 
    Parallel Tempering. The algorithm runs multiple replicas of the system at different
    temperatures and periodically attempts to swap configurations between adjacent
    temperature levels.

    Attributes
    ----------
    x : np.ndarray
        Current configuration state for all walkers.
    xmin : np.ndarray
        Minimum allowed values for parameters.
    xmax : np.ndarray
        Maximum allowed values for parameters.
    xstep : np.ndarray
        Step sizes for parameter updates.
    numsteps : int
        Total number of Monte Carlo steps to perform.
    numsteps_exchange : int
        Number of steps between exchange attempts.
    numsteps_thermalization : int
        Number of steps to discard for thermalization.
    fx : np.ndarray
        Current energy/objective function values.
    istep : int
        Current step number.
    nreplica : int
        Total number of replicas across all processes.
    Tindex : np.ndarray
        Temperature indices for current replicas.
    rep2T : np.ndarray
        Mapping from replica index to temperature index.
    T2rep : np.ndarray
        Mapping from temperature index to replica index.
    exchange_direction : bool
        Direction for attempting exchanges (alternates between True/False).
    """

    x: np.ndarray
    xmin: np.ndarray
    xmax: np.ndarray
    # xunit: np.ndarray
    xstep: np.ndarray

    numsteps: int
    numsteps_exchange: int

    fx: np.ndarray
    istep: int
    nreplica: int
    Tindex: np.ndarray
    rep2T: np.ndarray
    T2rep: np.ndarray

    exchange_direction: bool

    def __init__(
        self,
        info: odatse.Info,
        runner: odatse.Runner = None,
        run_mode: str = "initial",
        mpicomm: Optional["MPI.Comm"] = None,
    ) -> None:
        """
        Initialize the Algorithm class.

        Parameters
        ----------
        info : odatse.Info
            Information object containing algorithm parameters.
        runner : odatse.Runner, optional
            Runner object for executing the algorithm.
        run_mode : str, optional
            Mode to run the algorithm in, by default "initial".
        mpicomm : MPI.Comm
            MPI communicator to use for parallelization.
            If not provided, the default MPI communicator (MPI.COMM_WORLD) will be used if mpi4py is installed.
        """
        time_sta = time.perf_counter()

        info_exchange = info.algorithm["exchange"]
        nwalkers = info_exchange.get("nreplica_per_proc", 1)

        super().__init__(
            info=info,
            runner=runner,
            nwalkers=nwalkers,
            run_mode=run_mode,
            mpicomm=mpicomm,
        )

        self.nreplica = self.mpisize * self.nwalkers
        self.input_as_beta, self.betas = read_Ts(info_exchange, numT=self.nreplica)

        self.numsteps = info_exchange["numsteps"]
        self.numsteps_exchange = info_exchange["numsteps_exchange"]
        self.numsteps_thermalization = info_exchange.get("numsteps_thermalization", int(0.1*self.numsteps))

        self.export_combined_files = info_exchange.get("export_combined_files", False)
        self.separate_T = info_exchange.get("separate_T", True)

        time_end = time.perf_counter()
        self.timer["init"]["total"] = time_end - time_sta

    def _initialize(self) -> None:
        """
        Initialize the algorithm parameters and state.
        """
        # Initialize base class first
        super()._initialize()

        # Set up temperature indices for each walker
        # Each process handles a contiguous block of temperature indices
        # based on its rank and number of walkers
        self.Tindex = np.arange(
            self.mpirank * self.nwalkers, (self.mpirank + 1) * self.nwalkers
        )

        # Initialize mappings between replica and temperature indices
        # Initially, replica i has temperature i
        self.rep2T = np.arange(self.nreplica)  # Maps replica index -> temperature index
        self.T2rep = np.arange(self.nreplica)  # Maps temperature index -> replica index

        # Initialize exchange direction - alternates between True/False
        # to ensure all adjacent pairs get chance to exchange
        self.exchange_direction = True
        self.istep = 0

        self._show_parameters()

    def _run(self) -> None:
        """
        Run the algorithm.
        """
        # Validate run mode is set
        if self.mode is None:
            raise RuntimeError("mode unset")

        # Determine whether to restore RNG state from checkpoint
        restore_rng = not self.mode.endswith("-resetrand")

        # Initialize or restore simulation state based on mode
        if self.mode.startswith("init"):
            self._initialize()
        elif self.mode.startswith("resume"):
            self._load_state(self.checkpoint_file, mode="resume", restore_rng=restore_rng)
        elif self.mode.startswith("continue"):
            self._load_state(self.checkpoint_file, mode="continue", restore_rng=restore_rng)
        else:
            raise RuntimeError("unknown mode {}".format(self.mode))

        # Get current beta (inverse temperature) values for each replica
        beta = self.betas[self.Tindex]

        # Set up output file writers
        write_mode = "w" if self.mode.startswith("init") else "a"
        item_list = [
            "step",
            "walker",
            ("beta" if self.input_as_beta else "T"),
            "fx",
            *self.label_list,
        ]
        # Create writers for both trial moves and accepted results
        fp_trial = DataWriter("trial.txt", mode=write_mode, item_list=item_list, combined=self.export_combined_files)
        fp_result = DataWriter("result.txt", mode=write_mode, item_list=item_list, combined=self.export_combined_files)
        self._set_writer(fp_trial, fp_result)

        # For new runs, evaluate initial configuration
        if self.mode.startswith("init"):
            self.fx = self._evaluate(self.state)
            self._write_result(fp_trial)
            self._write_result(fp_result)
            self.istep += 1

            # Track best solution found
            minidx = np.argmin(self.fx)
            self.best_x = copy.copy(self.state.x[minidx, :])
            self.best_fx = np.min(self.fx[minidx])
            self.best_istep = 0
            self.best_iwalker = 0

        # Set up checkpointing intervals
        next_checkpoint_step = self.istep + self.checkpoint_steps
        next_checkpoint_time = time.time() + self.checkpoint_interval

        # Main simulation loop
        while self.istep < self.numsteps:
            # Attempt replica exchange periodically
            if self.istep % self.numsteps_exchange == 0:
                time_sta = time.perf_counter()
                if self.nreplica > 1:
                    self._exchange(self.exchange_direction)
                # Alternate exchange direction for next attempt
                if self.nreplica > 2:
                    self.exchange_direction = not self.exchange_direction
                time_end = time.perf_counter()
                self.timer["run"]["exchange"] += time_end - time_sta
                # Update beta values after exchange
                beta = self.betas[self.Tindex]

            # Perform local Monte Carlo updates
            self.local_update(beta)
            self.istep += 1

            # Handle checkpointing if enabled
            if self.checkpoint:
                time_now = time.time()
                if self.istep >= next_checkpoint_step or time_now >= next_checkpoint_time:
                    self._save_state(self.checkpoint_file)
                    next_checkpoint_step = self.istep + self.checkpoint_steps
                    next_checkpoint_time = time_now + self.checkpoint_interval

        # Clean up file handles
        fp_trial.close()
        fp_result.close()
        print("complete main process : rank {:08d}/{:08d}".format(self.mpirank, self.mpisize))

        # Save final state for possible continuation
        if self.checkpoint:
            self._save_state(self.checkpoint_file)

    def _exchange(self, direction: bool) -> None:
        """
        Attempt temperature exchanges between replicas.

        This method implements the core replica exchange logic, attempting to swap
        temperatures between adjacent replicas based on the Metropolis criterion:
        P(accept) = min(1, exp((β_j - β_i)(E_i - E_j)))

        Parameters
        ----------
        direction : bool
            If True, attempt exchanges between even-odd pairs.
            If False, attempt exchanges between odd-even pairs.
        """
        if self.nwalkers == 1:
            self.__exchange_single_walker(direction)
        else:
            self.__exchange_multi_walker(direction)

    def __exchange_single_walker(self, direction: bool) -> None:
        """
        Handle temperature exchanges for single walker per process case.

        This method implements the exchange logic when each process has only one walker,
        requiring MPI communication to coordinate exchanges between processes.

        Parameters
        ----------
        direction : bool
            If True, attempt exchanges between even-odd pairs.
            If False, attempt exchanges between odd-even pairs.
        """
        if self.mpisize > 1:
            self.mpicomm.Barrier()
        if direction:
            if self.Tindex[0] % 2 == 0:
                other_index = self.Tindex[0] + 1
                is_main = True
            else:
                other_index = self.Tindex[0] - 1
                is_main = False
        else:
            if self.Tindex[0] % 2 == 0:
                other_index = self.Tindex[0] - 1
                is_main = False
            else:
                other_index = self.Tindex[0] + 1
                is_main = True

        ibuf = np.zeros(1, dtype=np.int64)
        fbuf = np.zeros(1, dtype=np.float64)

        if 0 <= other_index < self.nreplica:
            other_rank = self.T2rep[other_index]
            if is_main:
                self.mpicomm.Recv(fbuf, source=other_rank, tag=1)
                other_fx = fbuf[0]
                beta = self.betas[self.Tindex[0]]
                other_beta = self.betas[self.Tindex[0] + 1]
                logp = (other_beta - beta) * (other_fx - self.fx[0])
                if logp >= 0.0 or self.rng.rand() < np.exp(logp):
                    ibuf[0] = self.Tindex
                    self.mpicomm.Send(ibuf, dest=other_rank, tag=2)
                    self.Tindex[0] += 1
                else:
                    ibuf[0] = self.Tindex + 1
                    self.mpicomm.Send(ibuf, dest=other_rank, tag=2)
            else:
                fbuf[0] = self.fx[0]
                self.mpicomm.Send(fbuf, dest=other_rank, tag=1)
                self.mpicomm.Recv(ibuf, source=other_rank, tag=2)
                self.Tindex[0] = ibuf[0]

        self.mpicomm.Barrier()
        if self.mpirank == 0:
            self.T2rep[self.Tindex[0]] = self.mpirank
            for other_rank in range(1, self.nreplica):
                self.mpicomm.Recv(ibuf, source=other_rank, tag=0)
                self.T2rep[ibuf[0]] = other_rank
        else:
            ibuf[0] = self.Tindex
            self.mpicomm.Send(ibuf, dest=0, tag=0)
        self.mpicomm.Bcast(self.T2rep, root=0)

    def __exchange_multi_walker(self, direction: bool) -> None:
        """
        Handle temperature exchanges for multiple walkers per process case.

        This method implements the exchange logic when each process has multiple walkers,
        requiring collective MPI operations to coordinate exchanges across all processes.

        Parameters
        ----------
        direction : bool
            If True, attempt exchanges between even-odd pairs.
            If False, attempt exchanges between odd-even pairs.
        """
        comm = self.mpicomm
        if self.mpisize > 1:
            fx_all = comm.allgather(self.fx)
            fx_all = np.array(fx_all).flatten()
        else:
            fx_all = self.fx

        rep2T_diff = []
        T2rep_diff = []

        for irep in range(
            self.mpirank * self.nwalkers, (self.mpirank + 1) * self.nwalkers
        ):
            iT = self.rep2T[irep]
            if iT % 2 != 0:
                continue
            jT = iT + 1 if direction else iT - 1
            if jT < 0 or jT == self.nreplica:
                continue
            jrep = self.T2rep[jT]
            fdiff = fx_all[jrep] - fx_all[irep]
            bdiff = self.betas[jT] - self.betas[iT]
            logp = fdiff * bdiff
            if logp >= 0.0 or self.rng.rand() < np.exp(logp):
                rep2T_diff.append((irep, jT))  # this means self.rep2T[irep] = jT
                rep2T_diff.append((jrep, iT))
                T2rep_diff.append((iT, jrep))
                T2rep_diff.append((jT, irep))

        if self.mpisize > 1:
            rep2T_diff = comm.allgather(rep2T_diff)
            rep2T_diff = list(itertools.chain.from_iterable(rep2T_diff))  # flatten
            T2rep_diff = comm.allgather(T2rep_diff)
            T2rep_diff = list(itertools.chain.from_iterable(T2rep_diff))  # flatten

        for diff in rep2T_diff:
            self.rep2T[diff[0]] = diff[1]
        for diff in T2rep_diff:
            self.T2rep[diff[0]] = diff[1]
        self.Tindex = self.rep2T[
            self.mpirank * self.nwalkers : (self.mpirank + 1) * self.nwalkers
        ]

    def _prepare(self) -> None:
        """
        Prepare the algorithm for execution.
        """
        self.timer["run"]["submit"] = 0.0
        self.timer["run"]["exchange"] = 0.0

    def _post(self) -> Dict:
        """
        Post-process the results of the algorithm.
        """
        # Separate results by temperature if requested
        if self.separate_T and not self.export_combined_files:
            if self.mpirank == 0:
                print(f"start separateT {self.mpirank}")
                sys.stdout.flush()

            # Convert beta to temperature if needed
            Ts = self.betas if self.input_as_beta else 1.0 / self.betas

            # Organize results by temperature
            separateT(
                Ts=Ts,
                nwalkers=self.nwalkers,
                output_dir=self.output_dir,
                comm=self.mpicomm,
                use_beta=self.input_as_beta,
                buffer_size=10000,
            )
            calculate_statistics_from_separated_files(
                Ts=Ts,
                output_dir=self.output_dir,
                thermalization_steps=self.numsteps_thermalization,
                comm=self.mpicomm,
            )

        # Gather best results from all processes
        if self.mpisize > 1:
            # NOTE:
            # ``gather`` seems not to work with many processes (say, 32) in some MPI implementation.
            # ``Gather`` and ``allgather`` seem to work fine.
            # Since the performance is not so important here, we use ``allgather`` for simplicity.
            best_fx = self.mpicomm.allgather(self.best_fx)
            best_x = self.mpicomm.allgather(self.best_x)
            best_istep = self.mpicomm.allgather(self.best_istep)
            best_iwalker = self.mpicomm.allgather(self.best_iwalker)
        else:
            best_fx = [self.best_fx]
            best_x = [self.best_x]
            best_istep = [self.best_istep]
            best_iwalker = [self.best_iwalker]

        # Find process with best overall solution
        best_rank = np.argmin(best_fx)

        # Write best result to file (rank 0 only)
        if self.mpirank == 0:
            with open("best_result.txt", "w") as f:
                f.write(f"nprocs = {self.nreplica}\n")
                f.write(f"rank = {best_rank}\n")
                f.write(f"step = {best_istep[best_rank]}\n")
                f.write(f"walker = {best_iwalker[best_rank]}\n")
                f.write(f"fx = {best_fx[best_rank]}\n")
                for label, x in zip(self.label_list, best_x[best_rank]):
                    f.write(f"{label} = {x}\n")
            # Print summary to stdout
            print("Best Result:")
            print(f"  rank = {best_rank}")
            print(f"  step = {best_istep[best_rank]}")
            print(f"  walker = {best_iwalker[best_rank]}")
            print(f"  fx = {best_fx[best_rank]}")
            for label, x in zip(self.label_list, best_x[best_rank]):
                print(f"  {label} = {x}")

        # Return best solution information
        return {
            "x": best_x[best_rank],
            "fx": best_fx[best_rank],
            "nprocs": self.nreplica,
            "rank": best_rank,
            "step": best_istep[best_rank],
            "walker": best_iwalker[best_rank],
        }

    def _save_state(self, filename) -> None:
        """
        Save the current algorithm state to a checkpoint file.

        Saves all necessary data to resume the simulation later, including:
          - MPI configuration
          - Random number generator state
          - Timer information
          - Current configurations and energies
          - Best solutions found
          - Replica exchange state information

        Parameters
        ----------
        filename : str
            Path to the checkpoint file to write.
        """
        data = {
            #-- _algorithm
            "mpisize": self.mpisize,
            "mpirank": self.mpirank,
            "rng": self.rng.get_state(),
            "timer": self.timer,
            "info": self.info,
            #-- montecarlo
            "x": self.state,
            "fx": self.fx,
            #"inode": self.inode,
            "istep": self.istep,
            "best_x": self.best_x,
            "best_fx": self.best_fx,
            "best_istep": self.best_istep,
            "best_iwalker": self.best_iwalker,
            "naccepted": self.naccepted,
            "ntrial": self.ntrial,
            #-- exchange
            "nreplica": self.nreplica,
            "Tindex": self.Tindex,
            "rep2T": self.rep2T,
            "T2rep": self.T2rep,
            "exchange_direction": self.exchange_direction,
        }
        self._save_data(data, filename)

    def _load_state(self, filename, mode="resume", restore_rng=True):
        """
        Load algorithm state from a checkpoint file.

        Restores all necessary data to resume a previous simulation run.

        Parameters
        ----------
        filename : str
            Path to the checkpoint file to read.
        mode : str, optional
            Loading mode - either "resume" or "continue", by default "resume".
        restore_rng : bool, optional
            Whether to restore the random number generator state, by default True.
            Set to False to use a fresh RNG state.

        Raises
        ------
        AssertionError
            If loaded state doesn't match current MPI configuration.
        """
        data = self._load_data(filename)
        if not data:
            print("ERROR: Load status file failed")
            sys.exit(1)

        # -- _algorithm
        assert self.mpisize == data["mpisize"]
        assert self.mpirank == data["mpirank"]

        if restore_rng:
            self.rng = np.random.RandomState()
            self.rng.set_state(data["rng"])
        self.timer = data["timer"]

        info = data["info"]
        self._check_parameters(info)

        # -- montecarlo
        self.state = data["x"]
        self.fx = data["fx"]
        # self.inode = data["inode"]

        self.istep = data["istep"]

        self.best_x = data["best_x"]
        self.best_fx = data["best_fx"]
        self.best_istep = data["best_istep"]
        self.best_iwalker = data["best_iwalker"]

        self.naccepted = data["naccepted"]
        self.ntrial = data["ntrial"]

        # -- exchange
        assert self.nreplica == data["nreplica"]

        self.Tindex = data["Tindex"]
        self.rep2T = data["rep2T"]
        self.T2rep = data["T2rep"]
        self.exchange_direction = data["exchange_direction"]

        # -- restore rng state in statespace
        self.statespace.rng = self.rng



# SPDX-License-Identifier: MPL-2.0
#
# ODAT-SE -- an open framework for data analysis
# Copyright (C) 2020- The University of Tokyo
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import time
import odatse
import numpy as np

# type hints
from pathlib import Path
from typing import List, Dict, Any, Optional

# Parameters
# ----------
# [runner.log]
#   interval
#   filename
#   write_input
#   write_result

class Logger:
    """
    Logger class to handle logging of calls, elapsed time, and optionally input and result data.
    """

    logfile: Path
    buffer_size: int
    buffer: List[str]
    num_calls: int
    time_start: float
    time_previous: float
    to_write_result: bool
    to_write_input: bool

    def __init__(self, info: Optional[odatse.Info] = None,
                 *,
                 buffer_size: int = 0,
                 filename: str = "runner.log",
                 write_input: bool = False,
                 write_result: bool = False,
                 params: Optional[Dict[str,Any]] = None,
                 **rest) -> None:
        """
        Initialize the Logger.

        Parameters
        ----------
        info : Info, optional
            Information object containing logging parameters.
        buffer_size : int
            Size of the buffer before writing to the log file.
        filename : str
            Name of the log file.
        write_input : bool
            Flag to indicate if input should be logged.
        write_result : bool
            Flag to indicate if result should be logged.
        params : Dict[str,Any]], optional
            Additional parameters for logging.
        **rest
            Additional keyword arguments.
        """
        if info is not None:
            info_log = info.runner.get("log", {})
        else:
            info_log = params

        self.buffer_size = info_log.get("interval", buffer_size)
        self.filename = info_log.get("filename", filename)
        self.to_write_input = info_log.get("write_input", write_input)
        self.to_write_result = info_log.get("write_result", write_result)

        self.time_start = time.perf_counter()
        self.time_previous = self.time_start
        self.num_calls = 0
        self.buffer = []

    def is_active(self) -> bool:
        """
        Check if logging is active.

        Returns
        -------
        bool
            True if logging is active, False otherwise.
        """
        return self.buffer_size > 0

    def prepare(self, proc_dir: Path) -> None:
        """
        Prepare the log file for writing.

        Parameters
        ----------
        proc_dir : Path
            Directory where the log file will be created.
        """
        if not self.is_active():
            return

        self.logfile = proc_dir / self.filename
        if self.logfile.exists():
            self.logfile.unlink()

        with open(self.logfile, "w") as f:
            f.write("# $1: num_calls\n")
            f.write("# $2: elapsed time from last call\n")
            f.write("# $3: elapsed time from start\n")
            if self.to_write_result:
                f.write("# $4: result\n")
            if self.to_write_input:
                f.write("# ${}-: input\n".format(5 if self.to_write_result else 4))
            f.write("\n")

    def count(self, x: np.ndarray, args, result: float) -> None:
        """
        Log a call with input and result data.

        Parameters
        ----------
        x : np.ndarray
            Input data.
        args :
            Additional arguments.
        result : float
            Result data.
        """
        if not self.is_active():
            return

        self.num_calls += 1

        t = time.perf_counter()

        fields = []
        fields.append(str(self.num_calls).ljust(6))
        fields.append("{:.6f}".format(t - self.time_previous))
        fields.append("{:.6f}".format(t - self.time_start))
        if self.to_write_result:
            fields.append(result)
        if self.to_write_input:
            for val in x:
                fields.append(val)
        self.buffer.append(" ".join(map(str, fields)) + "\n")

        self.time_previous = t

        if len(self.buffer) >= self.buffer_size:
            self.write()

    def write(self) -> None:
        """
        Write the buffered log entries to the log file.
        """
        if not self.is_active():
            return
        with open(self.logfile, "a") as f:
            for w in self.buffer:
                f.write(w)
        self.buffer.clear()
